import slay


class SplitText(slay.ProcessorBase):

    default_config = slay.Config(
        image=slay.Image()
        .pip_requirements_file(slay.make_abs_path_here("../requirements.txt"))
        .pip_requirements(["numpy"])
    )

    async def run(self, data: str, num_partitions: int) -> tuple[list[str], int]:
        import numpy as np

        parts = np.array_split(np.array(list(data)), num_partitions)
        return ["".join(part) for part in parts], 123



import math

import pydantic
import slay
from user_package import shared_processor

IMAGE_COMMON = slay.Image().pip_requirements_file(
    slay.make_abs_path_here("requirements.txt")
)


class GenerateData(slay.ProcessorBase):

    default_config = slay.Config(image=IMAGE_COMMON)

    def run(self, length: int) -> str:
        template = "erodfd"
        repetitions = int(math.ceil(length / len(template)))
        return (template * repetitions)[:length]


class DummyUserConfig(pydantic.BaseModel):
    multiplier: int


class TextReplicator(slay.ProcessorBase[DummyUserConfig]):

    default_config = slay.Config(
        image=IMAGE_COMMON,
        user_config=DummyUserConfig(multiplier=2),
    )

    def run(self, data: str) -> str:
        if len(data) > 30:
            raise ValueError(f"This input is too long: {len(data)}.")
        return data * self.user_config.multiplier


class TextToNum(slay.ProcessorBase):
    default_config = slay.Config(image=IMAGE_COMMON)

    def __init__(
        self,
        context: slay.Context = slay.provide_context(),
        replicator: TextReplicator = slay.provide(TextReplicator),
    ) -> None:
        super().__init__(context)
        self._replicator = replicator

    def run(self, data: str) -> int:
        number = 0
        generated_text = self._replicator.run(data)
        for char in generated_text:
            number += ord(char)

        return number


class Workflow(slay.ProcessorBase):
    default_config = slay.Config(image=IMAGE_COMMON)

    def __init__(
        self,
        context: slay.Context = slay.provide_context(),
        data_generator: GenerateData = slay.provide(GenerateData),
        splitter: shared_processor.SplitText = slay.provide(shared_processor.SplitText),
        text_to_num: TextToNum = slay.provide(TextToNum),
    ) -> None:
        super().__init__(context)
        self._data_generator = data_generator
        self._data_splitter = splitter
        self._text_to_num = text_to_num

    async def run(self, length: int, num_partitions: int) -> tuple[int, str, int]:
        data = self._data_generator.run(length)
        text_parts, number = await self._data_splitter.run(data, num_partitions)
        value = 0
        for part in text_parts:
            value += self._text_to_num.run(part)
        return value, data, number



import os
from itertools import count

import build_engine_utils
from builder.types import TrussTRTLLMConfiguration
from constants import (
    GRPC_SERVICE_PORT,
    HF_AUTH_KEY_CONSTANT,
    HTTP_SERVICE_PORT,
    TOKENIZER_KEY_CONSTANT,
)
from schema import ModelInput
from transformers import AutoTokenizer
from triton_client import TritonClient, TritonServer
from utils import execute_command

DEFAULT_MAX_TOKENS = 500
DEFAULT_MAX_NEW_TOKENS = 500


class Model:
    def __init__(self, data_dir, config, secrets):
        self._data_dir = data_dir
        self._config = config
        self._secrets = secrets
        self._request_id_counter = count(start=1)
        self.triton_client = None
        self.triton_server = None
        self.tokenizer = None
        self.uses_openai_api = None

    def load(self):
        execute_command(["ldconfig"])
        trtllm_config = TrussTRTLLMConfiguration(**self._config.get("trt_llm", {}))
        self.uses_openai_api = "openai-compatible" in self._config.get(
            "model_metadata", {}
        ).get("tags", [])
        hf_access_token = None
        if "hf_access_token" in self._secrets._base_secrets.keys():
            hf_access_token = self._secrets["hf_access_token"]

        # The underlying engine build call is idempotent, so we can call it in the load
        # even if the engine is already built. The engine build call checks to see if
        # there are .engine files in the destination directory and skips the build if
        # they are present.
        if trtllm_config.requires_build:
            build_engine_utils.build_engine_from_config_args(
                truss_trtllm_configuration=trtllm_config,
                checkpoint_dir_path=None,
                dst=self._data_dir,
            )

        self.triton_server = TritonServer(
            grpc_port=GRPC_SERVICE_PORT,
            http_port=HTTP_SERVICE_PORT,
        )

        if not trtllm_config.requires_build:
            engine_repository_path = trtllm_config.serve.engine_repository
            tokenizer_repository = trtllm_config.serve.tokenizer_repository
            tensor_parallel_count = trtllm_config.serve.tensor_parallel_count
            pipeline_parallel_count = trtllm_config.serve.pipeline_parallel_count
        else:
            # If this model required a build, the engine live inside the data_dir
            engine_repository_path = self._data_dir
            tokenizer_repository = trtllm_config.build.huggingface_ckpt_repository
            tensor_parallel_count = trtllm_config.build.tensor_parallel_count
            pipeline_parallel_count = trtllm_config.build.pipeline_parallel_count

        world_size = tensor_parallel_count * pipeline_parallel_count

        self.triton_server.create_model_repository(
            truss_data_dir=self._data_dir,
            engine_repository_path=engine_repository_path,
            huggingface_auth_token=hf_access_token,
        )

        env = {}
        if hf_access_token:
            env[HF_AUTH_KEY_CONSTANT] = hf_access_token
        env[TOKENIZER_KEY_CONSTANT] = tokenizer_repository

        self.triton_server.start(
            world_size=world_size,
            env=env,
        )

        self.triton_client = TritonClient(
            grpc_service_port=GRPC_SERVICE_PORT,
        )

        self.tokenizer = AutoTokenizer.from_pretrained(
            tokenizer_repository, token=hf_access_token
        )
        self.eos_token_id = self.tokenizer.eos_token_id

    async def predict(self, model_input):
        if "messages" not in model_input and "prompt" not in model_input:
            raise ValueError("Prompt or messages must be provided")

        model_input.setdefault("max_tokens", DEFAULT_MAX_TOKENS)
        model_input.setdefault("max_new_tokens", DEFAULT_MAX_NEW_TOKENS)
        model_input["request_id"] = str(os.getpid()) + str(
            next(self._request_id_counter)
        )
        model_input["eos_token_id"] = self.eos_token_id

        if "messages" in model_input:
            messages = model_input.pop("messages")
            if self.uses_openai_api and "prompt" not in model_input:
                model_input["prompt"] = self.tokenizer.apply_chat_template(
                    messages, tokenize=False
                )

        self.triton_client.start_grpc_stream()
        model_input = ModelInput(**model_input)
        result_iterator = self.triton_client.infer(model_input)

        async def generate():
            async for result in result_iterator:
                yield result

        if model_input.stream:
            return generate()
        else:
            if self.uses_openai_api:
                return "".join(generate())
            else:
                return {"text": "".join(generate())}



from pathlib import Path
from typing import Optional

from builder.types import TrussTRTLLMConfiguration


def build_engine_from_config_args(
    truss_trtllm_configuration: TrussTRTLLMConfiguration,
    dst: Path,
    checkpoint_dir_path: Optional[Path] = None,
):
    # NOTE: These are provided by the underlying base image
    # TODO(Abu): Remove this when we have a better way of handling this
    from builder.main import build_engine

    build_engine(
        engine_configuration=truss_trtllm_configuration,
        engine_serialization_path=dst,
        # If checkpoint_dir_path is provided, we'll look there for the
        # weight files. If not, we will attempt to use the `huggingface_ckpt_repository`
        # key in the `truss_trtllm_configuration` to download the weights.
        checkpoint_dir_path=checkpoint_dir_path,
    )
    return dst



from pathlib import Path
from typing import Any, Callable, List, Optional

import yaml
from truss.patch.calc_patch import calc_truss_patch, calc_unignored_paths
from truss.patch.signature import calc_truss_signature
from truss.templates.control.control.helpers.types import (
    Action,
    ConfigPatch,
    EnvVarPatch,
    ExternalDataPatch,
    ModelCodePatch,
    PackagePatch,
    Patch,
    PatchType,
    PythonRequirementPatch,
    SystemPackagePatch,
)
from truss.truss_config import TrussConfig
from truss.truss_handle import TrussHandle


def test_calc_truss_patch_unsupported(custom_model_truss_dir: Path):
    prev_sign = calc_truss_signature(custom_model_truss_dir)

    # Unsupported directory should result in no patches
    (custom_model_truss_dir / "data").touch()
    patches = calc_truss_patch(custom_model_truss_dir, prev_sign)
    assert len(patches) == 0

    # Changes under unsupported directory should return None to reflect
    # inability to calculate patch.
    (custom_model_truss_dir / "data" / "dummy").touch()
    patches = calc_truss_patch(custom_model_truss_dir, prev_sign)
    assert patches is None


def test_calc_truss_patch_add_file(custom_model_truss_dir: Path):
    prev_sign = calc_truss_signature(custom_model_truss_dir)
    with (custom_model_truss_dir / "model" / "dummy").open("w") as file:
        file.write("content")
    patches = calc_truss_patch(custom_model_truss_dir, prev_sign)

    assert len(patches) == 1
    patch = patches[0]
    assert patch == Patch(
        type=PatchType.MODEL_CODE,
        body=ModelCodePatch(
            action=Action.ADD,
            path="dummy",
            content="content",
        ),
    )


def test_calc_truss_patch_add_under_new_directory(custom_model_truss_dir: Path):
    prev_sign = calc_truss_signature(custom_model_truss_dir)
    new_dir = custom_model_truss_dir / "model" / "dir"
    new_dir.mkdir()
    (new_dir / "dummy").touch()
    patches = calc_truss_patch(custom_model_truss_dir, prev_sign)

    assert len(patches) == 1
    patch = patches[0]
    assert patch == Patch(
        type=PatchType.MODEL_CODE,
        body=ModelCodePatch(
            action=Action.ADD,
            path="dir/dummy",
            content="",
        ),
    )


def test_calc_truss_patch_remove_file(custom_model_truss_dir: Path):
    prev_sign = calc_truss_signature(custom_model_truss_dir)
    (custom_model_truss_dir / "model" / "model.py").unlink()
    patches = calc_truss_patch(custom_model_truss_dir, prev_sign)

    assert len(patches) == 1
    patch = patches[0]
    assert patch == Patch(
        type=PatchType.MODEL_CODE,
        body=ModelCodePatch(
            action=Action.REMOVE,
            path="model.py",
        ),
    )


def test_calc_truss_patch_update_file(custom_model_truss_dir: Path):
    prev_sign = calc_truss_signature(custom_model_truss_dir)
    new_model_file_content = """
    class Model:
        pass
    """
    with (custom_model_truss_dir / "model" / "model.py").open("w") as model_file:
        model_file.write(new_model_file_content)

    patches = calc_truss_patch(custom_model_truss_dir, prev_sign)

    assert len(patches) == 1
    patch = patches[0]
    assert patch == Patch(
        type=PatchType.MODEL_CODE,
        body=ModelCodePatch(
            action=Action.UPDATE, path="model.py", content=new_model_file_content
        ),
    )


def test_calc_truss_ignore_pycache(custom_model_truss_dir: Path):
    prev_sign = calc_truss_signature(custom_model_truss_dir)
    top_pycache_path = custom_model_truss_dir / "__pycache__"
    top_pycache_path.mkdir()
    (top_pycache_path / "bla.pyc").touch()
    model_pycache_path = custom_model_truss_dir / "model" / "__pycache__"
    model_pycache_path.mkdir()
    (model_pycache_path / "foo.pyo").touch()

    patches = calc_truss_patch(
        custom_model_truss_dir,
        prev_sign,
    )
    assert len(patches) == 0


def test_calc_truss_ignore_pycache_existing(custom_model_truss_dir: Path):
    # If __pycache__ existed before and there are no changes, there should be no
    # patches.
    top_pycache_path = custom_model_truss_dir / "__pycache__"
    top_pycache_path.mkdir()
    (top_pycache_path / "bla.pyc").touch()
    model_pycache_path = custom_model_truss_dir / "model" / "__pycache__"
    model_pycache_path.mkdir()
    (model_pycache_path / "foo.pyo").touch()
    sign = calc_truss_signature(custom_model_truss_dir)
    patches = calc_truss_patch(
        custom_model_truss_dir,
        sign,
    )
    assert len(patches) == 0


def test_calc_truss_ignore_changes_outside_patch_relevant_dirs(
    custom_model_truss_dir: Path,
):
    prev_sign = calc_truss_signature(custom_model_truss_dir)
    top_pycache_path = custom_model_truss_dir / "__pycache__"
    top_pycache_path.mkdir()
    (top_pycache_path / "README.md").touch()
    git_dir = custom_model_truss_dir / ".git"
    git_dir.mkdir()
    (git_dir / "dummy").touch()

    patches = calc_truss_patch(
        custom_model_truss_dir,
        prev_sign,
    )
    assert len(patches) == 0

    # Removing should also be ignored
    new_sign = calc_truss_signature(custom_model_truss_dir)
    (git_dir / "dummy").unlink()
    patches = calc_truss_patch(
        custom_model_truss_dir,
        new_sign,
    )
    assert len(patches) == 0


def test_calc_config_patches_add_python_requirement(custom_model_truss_dir: Path):
    patches = _apply_config_change_and_calc_patches(
        custom_model_truss_dir,
        lambda config: config.requirements.append("requests==1.0.0"),
    )
    assert len(patches) == 2
    assert patches == [
        Patch(
            type=PatchType.CONFIG,
            body=ConfigPatch(
                action=Action.UPDATE,
                config=yaml.safe_load((custom_model_truss_dir / "config.yaml").open()),
            ),
        ),
        Patch(
            type=PatchType.PYTHON_REQUIREMENT,
            body=PythonRequirementPatch(
                action=Action.ADD,
                requirement="requests==1.0.0",
            ),
        ),
    ]


def test_calc_truss_patch_add_package(custom_model_truss_dir: Path):
    prev_sign = calc_truss_signature(custom_model_truss_dir)
    new_dir = custom_model_truss_dir / "packages" / "dir"
    new_dir.mkdir()
    (new_dir / "dummy").touch()
    patches = calc_truss_patch(custom_model_truss_dir, prev_sign)

    assert len(patches) == 1
    patch = patches[0]
    assert patch == Patch(
        type=PatchType.PACKAGE,
        body=PackagePatch(
            action=Action.ADD,
            path="dir/dummy",
            content="",
        ),
    )


def test_calc_truss_patch_remove_package(
    custom_model_truss_dir_with_bundled_packages: Path,
):
    prev_sign = calc_truss_signature(custom_model_truss_dir_with_bundled_packages)
    (
        custom_model_truss_dir_with_bundled_packages
        / "packages"
        / "test_package"
        / "test.py"
    ).unlink()
    patches = calc_truss_patch(custom_model_truss_dir_with_bundled_packages, prev_sign)

    assert len(patches) == 1
    patch = patches[0]
    assert patch == Patch(
        type=PatchType.PACKAGE,
        body=PackagePatch(
            action=Action.REMOVE,
            path="test_package/test.py",
        ),
    )


def test_calc_truss_patch_update_package(
    custom_model_truss_dir_with_bundled_packages: Path,
):
    prev_sign = calc_truss_signature(custom_model_truss_dir_with_bundled_packages)
    new_package_file_content = """X = 2"""
    with (
        custom_model_truss_dir_with_bundled_packages
        / "packages"
        / "test_package"
        / "test.py"
    ).open("w") as package_file:
        package_file.write(new_package_file_content)

    patches = calc_truss_patch(custom_model_truss_dir_with_bundled_packages, prev_sign)

    assert len(patches) == 1
    patch = patches[0]
    assert patch == Patch(
        type=PatchType.PACKAGE,
        body=PackagePatch(
            action=Action.UPDATE,
            path="test_package/test.py",
            content=new_package_file_content,
        ),
    )


def test_calc_config_patches_remove_python_requirement(custom_model_truss_dir: Path):
    patches = _apply_config_change_and_calc_patches(
        custom_model_truss_dir,
        config_pre_op=lambda config: config.requirements.append("requests==1.0.0"),
        config_op=lambda config: config.requirements.clear(),
    )
    assert len(patches) == 2
    assert patches == [
        Patch(
            type=PatchType.CONFIG,
            body=ConfigPatch(
                action=Action.UPDATE,
                config=yaml.safe_load((custom_model_truss_dir / "config.yaml").open()),
            ),
        ),
        Patch(
            type=PatchType.PYTHON_REQUIREMENT,
            body=PythonRequirementPatch(
                action=Action.REMOVE,
                requirement="requests",
            ),
        ),
    ]


def test_calc_config_patches_update_python_requirement(custom_model_truss_dir: Path):
    def update_requests_version(config: TrussConfig):
        config.requirements[0] = "requests==2.0.0"

    patches = _apply_config_change_and_calc_patches(
        custom_model_truss_dir,
        config_pre_op=lambda config: config.requirements.append("requests==1.0.0"),
        config_op=update_requests_version,
    )
    assert len(patches) == 2
    assert patches == [
        Patch(
            type=PatchType.CONFIG,
            body=ConfigPatch(
                action=Action.UPDATE,
                config=yaml.safe_load((custom_model_truss_dir / "config.yaml").open()),
            ),
        ),
        Patch(
            type=PatchType.PYTHON_REQUIREMENT,
            body=PythonRequirementPatch(
                action=Action.UPDATE,
                requirement="requests==2.0.0",
            ),
        ),
    ]


def test_calc_config_patches_add_remove_and_update_python_requirement(
    custom_model_truss_dir: Path,
):
    def config_pre_op(config: TrussConfig):
        config.requirements = [
            "requests==1.0.0",
            "jinja==4.0.0",
        ]

    def config_op(config: TrussConfig):
        config.requirements = [
            "requests==2.0.0",
            "numpy>=1.8",
        ]

    patches = _apply_config_change_and_calc_patches(
        custom_model_truss_dir,
        config_pre_op=config_pre_op,
        config_op=config_op,
    )
    assert len(patches) == 4
    assert patches[0] == Patch(
        type=PatchType.CONFIG,
        body=ConfigPatch(
            action=Action.UPDATE,
            config=yaml.safe_load((custom_model_truss_dir / "config.yaml").open()),
        ),
    )
    patches = patches[1:]
    patches.sort(key=lambda patch: patch.body.requirement)
    assert patches == [
        Patch(
            type=PatchType.PYTHON_REQUIREMENT,
            body=PythonRequirementPatch(
                action=Action.REMOVE,
                requirement="jinja",
            ),
        ),
        Patch(
            type=PatchType.PYTHON_REQUIREMENT,
            body=PythonRequirementPatch(
                action=Action.ADD,
                requirement="numpy>=1.8",
            ),
        ),
        Patch(
            type=PatchType.PYTHON_REQUIREMENT,
            body=PythonRequirementPatch(
                action=Action.UPDATE,
                requirement="requests==2.0.0",
            ),
        ),
    ]


def test_calc_config_patches_add_env_var(
    custom_model_truss_dir: Path,
):
    patches = _apply_config_change_and_calc_patches(
        custom_model_truss_dir,
        config_op=lambda config: config.environment_variables.update({"foo": "bar"}),
    )
    assert len(patches) == 2
    assert patches == [
        Patch(
            type=PatchType.CONFIG,
            body=ConfigPatch(
                action=Action.UPDATE,
                config=yaml.safe_load((custom_model_truss_dir / "config.yaml").open()),
            ),
        ),
        Patch(
            type=PatchType.ENVIRONMENT_VARIABLE,
            body=EnvVarPatch(
                action=Action.ADD,
                item={"foo": "bar"},
            ),
        ),
    ]


def test_calc_config_patches_add_remove_env_var(
    custom_model_truss_dir: Path,
):
    patches = _apply_config_change_and_calc_patches(
        custom_model_truss_dir,
        config_pre_op=lambda config: config.environment_variables.update(
            {"foo": "bar"}
        ),
        config_op=lambda config: config.environment_variables.clear(),
    )
    assert len(patches) == 2
    assert patches == [
        Patch(
            type=PatchType.CONFIG,
            body=ConfigPatch(
                action=Action.UPDATE,
                config=yaml.safe_load((custom_model_truss_dir / "config.yaml").open()),
            ),
        ),
        Patch(
            type=PatchType.ENVIRONMENT_VARIABLE,
            body=EnvVarPatch(
                action=Action.REMOVE,
                item={"foo": "bar"},
            ),
        ),
    ]


def test_calc_config_patches_add_system_package(custom_model_truss_dir: Path):
    patches = _apply_config_change_and_calc_patches(
        custom_model_truss_dir,
        lambda config: config.system_packages.append("curl"),
    )
    assert len(patches) == 2
    assert patches == [
        Patch(
            type=PatchType.CONFIG,
            body=ConfigPatch(
                action=Action.UPDATE,
                config=yaml.safe_load((custom_model_truss_dir / "config.yaml").open()),
            ),
        ),
        Patch(
            type=PatchType.SYSTEM_PACKAGE,
            body=SystemPackagePatch(
                action=Action.ADD,
                package="curl",
            ),
        ),
    ]


def test_calc_config_patches_remove_system_package(custom_model_truss_dir: Path):
    patches = _apply_config_change_and_calc_patches(
        custom_model_truss_dir,
        config_pre_op=lambda config: config.system_packages.append("curl"),
        config_op=lambda config: config.system_packages.clear(),
    )
    assert len(patches) == 2
    assert patches == [
        Patch(
            type=PatchType.CONFIG,
            body=ConfigPatch(
                action=Action.UPDATE,
                config=yaml.safe_load((custom_model_truss_dir / "config.yaml").open()),
            ),
        ),
        Patch(
            type=PatchType.SYSTEM_PACKAGE,
            body=SystemPackagePatch(
                action=Action.REMOVE,
                package="curl",
            ),
        ),
    ]


def test_calc_config_patches_add_and_remove_system_package(
    custom_model_truss_dir: Path,
):
    def config_pre_op(config: TrussConfig):
        config.system_packages = [
            "curl",
            "jq",
        ]

    def config_op(config: TrussConfig):
        config.system_packages = [
            "curl",
            "libsnd",
        ]

    patches = _apply_config_change_and_calc_patches(
        custom_model_truss_dir,
        config_pre_op=config_pre_op,
        config_op=config_op,
    )
    assert len(patches) == 3
    assert patches[0] == Patch(
        type=PatchType.CONFIG,
        body=ConfigPatch(
            action=Action.UPDATE,
            config=yaml.safe_load((custom_model_truss_dir / "config.yaml").open()),
        ),
    )
    patches = patches[1:]
    patches.sort(key=lambda patch: patch.body.package)
    assert patches == [
        Patch(
            type=PatchType.SYSTEM_PACKAGE,
            body=SystemPackagePatch(
                action=Action.REMOVE,
                package="jq",
            ),
        ),
        Patch(
            type=PatchType.SYSTEM_PACKAGE,
            body=SystemPackagePatch(
                action=Action.ADD,
                package="libsnd",
            ),
        ),
    ]


def test_calc_config_patches_toggle_apply_library_patches(custom_model_truss_dir: Path):
    def config_op(config: TrussConfig):
        config.apply_library_patches = False

    patches = _apply_config_change_and_calc_patches(custom_model_truss_dir, config_op)
    assert len(patches) == 1
    patch = patches[0]
    assert patch == Patch(
        type=PatchType.CONFIG,
        body=ConfigPatch(
            action=Action.UPDATE,
            config=yaml.safe_load((custom_model_truss_dir / "config.yaml").open()),
        ),
    )


def test_calc_config_patches_add_external_data(
    custom_model_external_data_access_tuple_fixture: Path,
):
    path, _ = custom_model_external_data_access_tuple_fixture
    th = TrussHandle(path)
    external_data = th.spec.config.external_data

    def config_op(config: TrussConfig):
        config.external_data = external_data

    def config_pre_op(config: TrussConfig):
        config.external_data = None

    patches = _apply_config_change_and_calc_patches(
        path,
        config_pre_op=config_pre_op,
        config_op=config_op,
    )
    assert len(patches) == 2

    assert patches == [
        Patch(
            type=PatchType.CONFIG,
            body=ConfigPatch(
                action=Action.UPDATE,
                config=yaml.safe_load((path / "config.yaml").open()),
            ),
        ),
        Patch(
            type=PatchType.EXTERNAL_DATA,
            body=ExternalDataPatch(action=Action.ADD, item=external_data.to_list()[0]),
        ),
    ]


def test_calc_config_patches_remove_external_data(
    custom_model_external_data_access_tuple_fixture: Path,
):
    path, _ = custom_model_external_data_access_tuple_fixture
    th = TrussHandle(path)
    external_data = th.spec.config.external_data

    def config_op(config: TrussConfig):
        config.external_data = None

    patches = _apply_config_change_and_calc_patches(
        path,
        config_op=config_op,
    )
    assert len(patches) == 2
    assert patches == [
        Patch(
            type=PatchType.CONFIG,
            body=ConfigPatch(
                action=Action.UPDATE,
                config=yaml.safe_load((path / "config.yaml").open()),
            ),
        ),
        Patch(
            type=PatchType.EXTERNAL_DATA,
            body=ExternalDataPatch(
                action=Action.REMOVE, item=external_data.to_list()[0]
            ),
        ),
    ]


def test_calc_unignored_paths():
    ignore_patterns = [
        ".mypy_cache/",
        "venv/",
        "*.tmp",
    ]

    root_relative_paths = {
        ".mypy_cache/should_ignore.json",
        "venv/bin/activate",
        "ignored_file.tmp",
        "config.yaml",
        "model/model.py",
    }

    unignored_paths = calc_unignored_paths(root_relative_paths, ignore_patterns)
    assert unignored_paths == {
        "config.yaml",
        "model/model.py",
    }


def _apply_config_change_and_calc_patches(
    custom_model_truss_dir: Path,
    config_op: Callable[[TrussConfig], Any],
    config_pre_op: Optional[Callable[[TrussConfig], Any]] = None,
) -> List[Patch]:
    def modify_config(op):
        config_path = custom_model_truss_dir / "config.yaml"
        config = TrussConfig.from_yaml(config_path)
        op(config)
        config.write_to_yaml_file(config_path)

    if config_pre_op is not None:
        modify_config(config_pre_op)

    prev_sign = calc_truss_signature(custom_model_truss_dir)
    modify_config(config_op)
    return calc_truss_patch(custom_model_truss_dir, prev_sign)



import json
import sys
from pathlib import Path

import pytest
import requests
from slay import definitions, framework
from truss.tests.test_testing_utilities_for_other_tests import ensure_kill_all

DEFAULT_LOG_ERROR = "Internal Server Error"


def _log_contains_error(line: dict, error: str, message: str):
    return (
        line["levelname"] == "ERROR"
        and line["message"] == message
        and error in line["exc_info"]
    )


def assert_logs_contain_error(logs: str, error: str, message=DEFAULT_LOG_ERROR):
    loglines = logs.splitlines()
    assert any(
        _log_contains_error(json.loads(line), error, message) for line in loglines
    )


@pytest.mark.integration
def test_workflow():
    with ensure_kill_all():
        root = Path(__file__).parent.parent.resolve()
        workflow_root = root / "test_data" / "workflow_text_to_num"

        sys.path.append(str(workflow_root))
        from workflow import Workflow

        sys.path.pop()

        options = definitions.DeploymentOptionsLocalDocker(
            workflow_name="integration-test"
        )
        entrypoint_service = framework.deploy_remotely(Workflow, options)
        predict_url = entrypoint_service.predict_url.replace(
            "host.docker.internal", "localhost"
        )

        response = requests.post(
            predict_url, json={"length": 30, "num_partitions": 3}, stream=True
        )
        print(response.content)
        assert response.status_code == 200
        assert response.json() == [6280, "erodfderodfderodfderodfderodfd", 123]

        # Test with errors.
        response = requests.post(
            predict_url, json={"length": 300, "num_partitions": 3}, stream=True
        )
        print(response)
        error = definitions.RemoteErrorDetail.parse_obj(response.json()["error"])
        print(error.format())
        assert response.status_code == 500



from unittest import mock

import pytest
import requests
from requests import Response
from truss.remote.baseten.api import BasetenApi
from truss.remote.baseten.error import ApiError


def mock_auth_service():
    auth_service = mock.Mock()
    auth_token = mock.Mock(headers=lambda: {"Authorization": "Api-Key token"})
    auth_service.authenticate.return_value = auth_token
    return auth_service


def mock_successful_response():
    response = Response()
    response.status_code = 200
    response.json = mock.Mock(return_value={"data": {"status": "success"}})
    return response


def mock_graphql_error_response():
    response = Response()
    response.status_code = 200
    response.json = mock.Mock(return_value={"errors": [{"message": "error"}]})
    return response


def mock_unsuccessful_response():
    response = Response()
    response.status_code = 400
    return response


def mock_create_model_version_response():
    response = Response()
    response.status_code = 200
    response.json = mock.Mock(
        return_value={"data": {"create_model_version_from_truss": {"id": "12345"}}}
    )
    return response


def mock_create_model_response():
    response = Response()
    response.status_code = 200
    response.json = mock.Mock(
        return_value={"data": {"create_model_from_truss": {"id": "12345"}}}
    )
    return response


@mock.patch("truss.remote.baseten.auth.AuthService")
@mock.patch("requests.post", return_value=mock_successful_response())
def test_post_graphql_query_success(mock_post, mock_auth_service):
    api_url = "https://test.com/api"
    api = BasetenApi(api_url, mock_auth_service)

    response_data = {"data": {"status": "success"}}

    result = api._post_graphql_query("sample_query_string")

    assert result == response_data


@mock.patch("truss.remote.baseten.auth.AuthService")
@mock.patch("requests.post", return_value=mock_graphql_error_response())
def test_post_graphql_query_error(mock_post, mock_auth_service):
    api_url = "https://test.com/api"
    api = BasetenApi(api_url, mock_auth_service)

    with pytest.raises(ApiError):
        api._post_graphql_query("sample_query_string")


@mock.patch("truss.remote.baseten.auth.AuthService")
@mock.patch("requests.post", return_value=mock_unsuccessful_response())
def test_post_requests_error(mock_post, mock_auth_service):
    api_url = "https://test.com/api"
    api = BasetenApi(api_url, mock_auth_service)
    with pytest.raises(requests.exceptions.HTTPError):
        api._post_graphql_query("sample_query_string")


@mock.patch("truss.remote.baseten.auth.AuthService")
@mock.patch("requests.post", return_value=mock_create_model_version_response())
def test_create_model_version_from_truss(mock_post, mock_auth_service):
    api_url = "https://test.com/api"
    api = BasetenApi(api_url, mock_auth_service)

    api.create_model_version_from_truss(
        "model_id",
        "s3key",
        "config_str",
        "semver_bump",
        "client_version",
        False,
        False,
        False,
        "deployment_name",
    )

    gql_mutation = mock_post.call_args[1]["data"]["query"]
    assert 'model_id: "model_id"' in gql_mutation
    assert 's3_key: "s3key"' in gql_mutation
    assert 'config: "config_str"' in gql_mutation
    assert 'semver_bump: "semver_bump"' in gql_mutation
    assert 'client_version: "client_version"' in gql_mutation
    assert "is_trusted: false" in gql_mutation
    assert "promote_after_deploy: false" in gql_mutation
    assert "scale_down_old_production: true" in gql_mutation
    assert 'name: "deployment_name"' in gql_mutation


@mock.patch("truss.remote.baseten.auth.AuthService")
@mock.patch("requests.post", return_value=mock_create_model_version_response())
def test_create_model_version_from_truss_does_not_send_deployment_name_if_not_specified(
    mock_post, mock_auth_service
):
    api_url = "https://test.com/api"
    api = BasetenApi(api_url, mock_auth_service)

    api.create_model_version_from_truss(
        "model_id",
        "s3key",
        "config_str",
        "semver_bump",
        "client_version",
        True,
        True,
        False,
        deployment_name=None,
    )

    gql_mutation = mock_post.call_args[1]["data"]["query"]
    assert 'model_id: "model_id"' in gql_mutation
    assert 's3_key: "s3key"' in gql_mutation
    assert 'config: "config_str"' in gql_mutation
    assert 'semver_bump: "semver_bump"' in gql_mutation
    assert 'client_version: "client_version"' in gql_mutation
    assert "is_trusted: true" in gql_mutation
    assert "promote_after_deploy: true" in gql_mutation
    assert "scale_down_old_production: true" in gql_mutation
    assert "name: " not in gql_mutation


@mock.patch("truss.remote.baseten.auth.AuthService")
@mock.patch("requests.post", return_value=mock_create_model_version_response())
def test_create_model_version_from_truss_does_not_scale_old_prod_to_zero_if_keep_previous_prod_settings(
    mock_post, mock_auth_service
):
    api_url = "https://test.com/api"
    api = BasetenApi(api_url, mock_auth_service)

    api.create_model_version_from_truss(
        "model_id",
        "s3key",
        "config_str",
        "semver_bump",
        "client_version",
        True,
        True,
        True,
        deployment_name=None,
    )

    gql_mutation = mock_post.call_args[1]["data"]["query"]
    assert 'model_id: "model_id"' in gql_mutation
    assert 's3_key: "s3key"' in gql_mutation
    assert 'config: "config_str"' in gql_mutation
    assert 'semver_bump: "semver_bump"' in gql_mutation
    assert 'client_version: "client_version"' in gql_mutation
    assert "is_trusted: true" in gql_mutation
    assert "promote_after_deploy: true" in gql_mutation
    assert "scale_down_old_production: false" in gql_mutation
    assert "name: " not in gql_mutation


@mock.patch("truss.remote.baseten.auth.AuthService")
@mock.patch("requests.post", return_value=mock_create_model_response())
def test_create_model_from_truss(mock_post, mock_auth_service):
    api_url = "https://test.com/api"
    api = BasetenApi(api_url, mock_auth_service)

    api.create_model_from_truss(
        "model_name",
        "s3key",
        "config_str",
        "semver_bump",
        "client_version",
        False,
        "deployment_name",
    )

    gql_mutation = mock_post.call_args[1]["data"]["query"]
    assert 'name: "model_name"' in gql_mutation
    assert 's3_key: "s3key"' in gql_mutation
    assert 'config: "config_str"' in gql_mutation
    assert 'semver_bump: "semver_bump"' in gql_mutation
    assert 'client_version: "client_version"' in gql_mutation
    assert "is_trusted: false" in gql_mutation
    assert 'version_name: "deployment_name"' in gql_mutation


@mock.patch("truss.remote.baseten.auth.AuthService")
@mock.patch("requests.post", return_value=mock_create_model_response())
def test_create_model_from_truss_does_not_send_deployment_name_if_not_specified(
    mock_post, mock_auth_service
):
    api_url = "https://test.com/api"
    api = BasetenApi(api_url, mock_auth_service)

    api.create_model_from_truss(
        "model_name",
        "s3key",
        "config_str",
        "semver_bump",
        "client_version",
        True,
        deployment_name=None,
    )

    gql_mutation = mock_post.call_args[1]["data"]["query"]
    assert 'name: "model_name"' in gql_mutation
    assert 's3_key: "s3key"' in gql_mutation
    assert 'config: "config_str"' in gql_mutation
    assert 'semver_bump: "semver_bump"' in gql_mutation
    assert 'client_version: "client_version"' in gql_mutation
    assert "is_trusted: true" in gql_mutation
    assert "version_name: " not in gql_mutation
